from .api import API
from .statemonitor import StateMonitorProcess
from expyrimenter import Executor, SSH
from time import sleep
import threading
import logging


class CloudStack:
    _id_cache = None

    def __init__(s, executor=None, api=None):
        if executor is None:
            executor = Executor()
        if api is None:
            api = API()

        s.executor = executor
        s._api = api
        s._logger = logging.getLogger('cloudstack')

        s._sm_lock = threading.Lock()
        s._sm_users = 0

    def get_states(s):
        vms = s._api.listVirtualMachines()['virtualmachine']
        return {vm['name']: vm['state'] for vm in vms}

    def get_id(s, name):
        cache = CloudStack._id_cache
        if cache is None or name not in cache:
            s.load_id_cache()

        return CloudStack._id_cache[name]

    def get_state(s, name):
        vm_id = s.get_id(name)
        vms = s._api.listVirtualMachines(id=vm_id)['virtualmachine']
        return vms[0]['state']

    def start(s, *names):
        names = s._ensure_lists(names)
        for name in names:
            s._logger.info('Starting %s' % name)
            vm_id = s.get_id(name)
            s._submit_sm_user(start_vm, 'start VM', name, vm_id)

    def stop(s, *names):
        names = s._ensure_lists(names)
        for name in names:
            vm_id = s.get_id(name)
            s.executor.run_fn(stop_vm, 'stop VM', name, vm_id)

    def get_deploy_params(s, name):
        vm_id = s.get_id(name)
        vm = s._api.listVirtualMachines(id=vm_id)['virtualmachine'][0]
        keys = ['serviceofferingid', 'templateid', 'zoneid']

        params = {}
        for key in keys:
            params[key] = vm[key]

        return params

    def deploy(s, params, **kwargs):
        if kwargs:
            params.update(kwargs)
        s._submit_sm_user(deploy_vm, 'deploy VM', params)

    def deploy_like(s, existent, new, **kwargs):
        params = s.get_deploy_params(existent)
        params['name'] = new
        s.deploy(params, **kwargs)

    def load_id_cache(s):
        vms = s._api.listVirtualMachines()['virtualmachine']
        CloudStack._id_cache = {vm['name']: vm['id'] for vm in vms}
        return CloudStack._id_cache

    def _submit_sm_user(s, fn, title, *args, **kwargs):
        with s._sm_lock:
            s._sm_users += 1
            StateMonitorProcess.start()

        states = StateMonitorProcess.get_states()
        args += (states,)
        future = s.executor.run_fn(fn, title, *args, **kwargs)
        future.add_done_callback(s._sm_user_done)

        return future

    def _sm_user_done(s, future):
        with s._sm_lock:
            s._sm_users -= 1
            if s._sm_users == 0:
                StateMonitorProcess.stop()

    def _ensure_lists(s, args):
        lists = []
        for arg in args:
            if isinstance(arg, list):
                lists += arg
            else:
                lists.append(arg)
        return lists


def stop_vm(name, vm_id):
    api = API()
    api.stopVirtualMachine(id=vm_id)
    msg = 'Sent %s stop request.' % name
    logging.getLogger('cloudstack').debug(msg)


def start_vm(name, vm_id, states):
    api = API()
    api.startVirtualMachine(id=vm_id)
    wait_ssh(name, states)
    msg = '%s is up.' % name
    logging.getLogger('cloudstack').info(msg)


def deploy_vm(params, states):
    api = API()
    api.deployVirtualMachine(**params)
    log = logging.getLogger('cloudstack')
    if 'startvm' in params and params['startvm'] is False:
        msg = '%s is deployed.' % params['name']
        log.debug(msg)
    else:
        wait_ssh(params['name'], states)
        msg = '%s is ready for SSH.' % params['name']
        log.info(msg)


def wait_ssh(name, states):
    wait_state(name, 'Running', states)
    SSH.await_availability(name)


def wait_state(name, state, states, interval=5):
    while True:
        if state == states.get(name):
            break
        sleep(interval)



from concurrent.futures import ThreadPoolExecutor
from . import Config
import concurrent.futures
import logging
from subprocess import CalledProcessError


class Executor:
    def __init__(s, max_workers=None):
        if max_workers is None:
            max_workers = int(Config('executor').get('max_concurrency'))

        s._executor = ThreadPoolExecutor(max_workers)
        s._future_runnables = {}  # for submitted runnables
        s._function_titles = {}     # for submitted functions
        s.results = []
        s._log = logging.getLogger('executor')

    def run_function(s, function, title=None, *args, **kwargs):
        """Submits a function to be run in parallel.

        If you only want to submit a function, use this method.
        It is not mandatory to call wait() or shutdown() later.
        """
        future = s._executor.submit(function, *args, **kwargs)
        s._function_titles[future] = title
        future.add_done_callback(s._done_function)

        return future

    def run(s, runnable, *args, **kwargs):
        """Submits Runnable objects to the PoolExector.

        Using Runnable objects, you have more control and verbosity.
        Useful when things go wrong (we know it always happens).
        If you are in a hurry, submit a function using :py:func:`run_fn`.
        """
        future = s._executor.submit(runnable.run, *args, **kwargs)
        s._future_runnables[future] = runnable
        future.add_done_callback(s._done_runnable)

        return future

    def _done_function(s, future):
        title = s._function_titles[future]
        s._done(future, title)
        del s._function_titles[future]

    def _done_runnable(s, future):
        runnable = s._future_runnables[future]
        title = runnable.title
        s._done(future, title)
        del s._future_runnables[future]

    def _done(s, future, title):
        if title is None:
            title = 'no given title'
        result = None

        ex = future.exception()
        if ex is None:
            s._log.debug('success:%s' % title)
            result = future.result()
        else:
            if type(ex) is CalledProcessError:
                msg = 'CalledProcessError:'
                if title != ex.cmd:
                    msg += '\n\tTitle   : %s' % title
                msg += '\n\tCmd     : %s' % ex.cmd
                msg += '\n\tReturned: %s' % ex.returncode
                msg += '\n\tOutput  : %s' % ex.output
                result = ex.output
            else:
                msg = 'exception:%s:%s' % (title, ex)
            s._log.error(msg)

        s.results.append(result)

    def wait(s):
        futures = list(s._future_runnables.keys())
        futures += list(s._function_titles.keys())

        concurrent.futures.wait(futures)

    def shutdown(s):
        s._executor.shutdown()
        s._future_runnables.clear()
        s._function_titles.clear()
        del s.results[:]



from . import API
from multiprocessing import Manager, Process
from time import sleep
import logging
import signal


class StateMonitor:
    _stop = False

    def __init__(self, proxy_states):
        self._logger = logging.getLogger('cloudstack.statemonitor')
        self._logger.info('started')

        self._proxy_states = proxy_states
        self._local_states = {}
        self._api = API()

    def monitor_states(self, interval=5):
        while not StateMonitor._stop:
            self._monitor_states_once()
            sleep(interval)
        self._logger.info('finished')

    @staticmethod
    def stop():
        StateMonitor._stop = True

    def _monitor_states_once(self):
        vms = self._api.listVirtualMachines()['virtualmachine']
        for vm in vms:
            self._update_state(vm['name'], vm['state'])

    def _update_state(self, k, v):
        if v != self._local_states.get(k):
            self._local_states[k] = v
            self._proxy_states[k] = v


def handler(signum, frame):
    StateMonitor.stop()


signal.signal(signal.SIGTERM, handler)


def monitor_states(states):
    sm = StateMonitor(states)
    sm.monitor_states()


# Ensure only one state monitor is running at most
class StateMonitorProcess:
    _mgr, _states, _process = None, None, None

    @staticmethod
    def start():
        if StateMonitorProcess._process is None:
            cls = StateMonitorProcess
            cls._mgr = Manager()
            cls._states = cls._mgr.dict()
            cls._process = Process(target=monitor_states, args=(cls._states,))
            cls._process.start()

    @staticmethod
    def stop():
        if StateMonitorProcess._process is not None:
            cls = StateMonitorProcess
            cls._process.terminate()
            cls._process.join()
            cls._states.clear()
            cls._mgr.shutdown()
            cls._mgr, cls._states, cls._process = None, None, None

    @staticmethod
    def get_states():
        return StateMonitorProcess._states



import logging
import requests
import json
from expyrimenter import Config
from os.path import getmtime, expanduser
import time


class Pushbullet:
    """ Requires ``[pushbullet]`` section with ``token`` value in config.
    Check :class:`expyrimenter.Config` for more details about configuration.
    """

    _url = 'https://api.pushbullet.com/v2/pushes'

    def __init__(self):
        self._log = logging.getLogger('pushbullet')

        token = Config('pushbullet').get('token')
        if token is None:
            msg = 'No pushbullet access token found. Check Config docs.'
            self._log.error(msg)
            raise AttributeError(msg)
        self._token = token

    def send_note(self, title='', body=''):
        """
        Sends a message using pushbullet *note* type.

        :param str title: Optional message title
        :param str body: Optional message body content
        """
        note = json.dumps({'type': 'note', 'title': title, 'body': body})
        headers = {'Content-Type': 'application/json'}
        res = requests.post(Pushbullet._url, data=note, headers=headers,
                            auth=(self._token, ''))

        if res.status_code == requests.codes.ok:
            self._log.info('Note sent to pushbullet')
        else:
            error_json = json.dumps(res.json(), ensure_ascii=False, indent=4)
            self._log.error('Error sending note to pushbullet:\n' +
                            error_json)

    def monitor_file(self, filename, max_mod_interval, title=None, body=None):
        """
        Sends a message if *filename* is not modified within *max_mod_interval*
        seconds. When a message is sent, it stops monitoring the file.
        Otherwise, it keeps running until it is killed.

        :param str filename: The filename to be monitored
        :param int max_mod_interval: If *filename* is not modified within
         *max_mod_interval* seconds, a message will be sent
        :param str title: Optional message title
        :param str body: Optional message body content
        """
        if title is None:
            title = filename
        if body is None:
            body = 'File not changed in %d seconds.' % max_mod_interval
        filename = expanduser(filename)

        note_sent = False
        while not note_sent:  # onde message is enough
            # time since last file modification
            time_since = time.time() - getmtime(filename)
            if time_since >= max_mod_interval:
                self._log.info('File %s not modified' % filename)
                self.send_note(title, body)
                note_sent = True
            else:
                # Wait until max_mod_interval after last modification
                sleep = max_mod_interval - time_since  # we know this is > 0
                time.sleep(sleep)



from expyrimenter import Shell
from time import sleep
import logging
import random


class SSH(Shell):
    """
    :param str params: shell SSH params (at least the hostname).
    :param str remote_cmd: Command to be run in remote host through SSH.
    :param str title: A title to be displayed in log outputs.
                      If None, the shell command will be shown.
    :param bool stdout: Whether or not to display standard output.
                        Default is *False*.
    :param bool stderr: Whether or not to display standard error.
                        Default is *True*.
    """

    def __init__(self, params, remote_cmd, title=None,
                 stdout=False, stderr=True):
        remote_cmd = self._redirect_outputs(remote_cmd, stdout, stderr)
        cmd = "ssh %s '%s'" % (params, remote_cmd)
        super().__init__(cmd, title, stdout, stderr)

    @staticmethod
    def await_availability(params, interval=5, max_rand=1):
        """Periodically tries SSH until it is successful.
        This function is very useful in cloud environmentes, because
        there can be a considerable amount of time after a VM is running
        and before SSH connections are available.

        :param str params: shell SSH params (at least the hostname).
        :param num interval: Time in seconds to wait before new trial.
        :param num max_rand: A float random number between 0 and ``max_rand``
                             will be added to ``interval``.
        """
        ssh = SSH(params, 'exit')
        log = logging.getLogger('ssh')
        while ssh.has_failed():
            sleep(random.uniform(0, max_rand))
            log.debug('Trying "ssh %s" again in %d + [0, %.2f) sec' %
                      (params, interval, max_rand))
            sleep(interval)



import logging
import subprocess
from subprocess import CalledProcessError
from .runnable import Runnable


class Shell(Runnable):
    """
    :param str cmd: Command with arguments to be run.
    :param str title: A title to be displayed in log outputs.
                      If None, :attr:`cmd` will be shown.
    :param bool stdout: Whether or not to display standard output.
                        Default is *False*.
    :param bool stderr: Whether or not to display standard error.
                        Default is *True*.
    """

    def __init__(self, cmd, title=None, stdout=False, stderr=True):
        self._cmd = self._redirect_outputs(cmd, stdout, stderr)
        self._title, self._stdout, self._stderr = title, stdout, stderr

    def _redirect_outputs(self, cmd, stdout, stderr):
        """
        Avoid transfering unnecessary output through the network.
        If local, keeps the output clean.
        """
        suffix = ''
        if not stdout:
            suffix += ' 1>/dev/null'
        if not stderr:
            suffix += ' 2>/dev/null'

        return cmd + suffix

    @property
    def command(self):
        """
        The command to be run in shell.

        :rtype: str
        """
        return self._cmd

    @property
    def title(self):
        """
        If the title is not set, returns :func:`command`.

        :rtype: str
        """
        if self._title is None:
            return self._cmd
        return self._title

    def run_pre(self):
        logger = logging.getLogger('shell')
        logger.debug('beginning:' + self.title)

    def run(self):
        """If the command exits 0, returns 0 or the stdout/stderr output.
        Otherwise, raises CalledProcessError.

        :return: shell return code (0) or output.
        :rtype: int or str
        :raises CalledProcessError: if return code is not 0.
        """
        self.run_pre()

        kwargs = {'shell': True, 'universal_newlines': True}
        if self._stderr:
            kwargs['stderr'] = subprocess.STDOUT

        if not (self._stdout or self._stderr):
            output = subprocess.check_call(self._cmd, **kwargs)
        else:
            output = subprocess.check_output(self._cmd, **kwargs).strip()

        self.run_pos()
        return output

    def was_successful(self):
        """Runs the command and returns whether the return code is 0.

        :return: whether the command was successful.
        :rtype: bool
        """
        try:
            self.run()
            return True
        except CalledProcessError:
            return False

    def has_failed(self):
        """Runs the command and returns whether the return code differs from 0.

        :return: whether the command has failed.
        :rtype: bool
        """
        return not self.was_successful()

    def run_pos(self):
        logger = logging.getLogger('shell')
        logger.debug('finished:' + self.title)



# By: Carlos Eduardo Moreira dos Santos, 2014.
# Based on: Kelcey Damage, 2012 & Kraig Amador, 2012
# Changes:
#   - ported to python3
#   - support for config.ini (url, key, secret)
#   - method calls work with no parameters
#   - pep8 compliance
#   - documentation

from expyrimenter import Config
from urllib.parse import quote_plus
from urllib.request import urlopen
from urllib.error import HTTPError
import base64
import hashlib
import hmac
import json
import logging


class SignedAPICall():
    def __init__(self, api_url, api_key, api_secret):
        self.url = api_url
        self.key = api_key
        self.secret = api_secret
        self._logger = logging.getLogger('cloudstack.api')

    def request(self, args):
        args['apiKey'] = self.key

        self.params = []
        self._sort_request(args)
        self._create_signature()
        self._build_post_request()

    def _sort_request(self, args):
        keys = sorted(args.keys())

        for key in keys:
            self.params.append(key + '=' + self._quote(args[key]))

    def _quote(self, value):
        if value is True:
            quoted = 'true'
        elif value is False:
            quoted = 'false'
        else:
            quoted = quote_plus(value)

        return quoted

    def _create_signature(self):
        self.query = '&'.join(self.params)
        digest = hmac.new(self.secret.encode(),
                          msg=self.query.lower().encode(),
                          digestmod=hashlib.sha1).digest()

        self.signature = base64.b64encode(digest)

    def _build_post_request(self):
        self.query += '&signature=' + quote_plus(self.signature)
        self.value = self.url + '?' + self.query


class API(SignedAPICall):
    def __init__(self):
        cfg = Config('cloudstack')
        super().__init__(cfg.get('url'), cfg.get('key'), cfg.get('secret'))

    def __getattr__(self, name):
        def handlerFunction(*args, **kwargs):
            if kwargs:
                return self._make_request(name, kwargs)
            if len(args) == 0:
                args = [{}]
            return self._make_request(name, args[0])
        return handlerFunction

    def _http_get(self, url):
        try:
            response = urlopen(url)
        except HTTPError:
            self._logger.error('Error getting "%s"' % url)
            raise

        return response.read()

    def _make_request(self, command, args):
        args['response'] = 'json'
        args['command'] = command
        self.request(args)
        data = self._http_get(self.value).decode()
        # The response is of the format {commandresponse: actual-data}
        key = command.lower() + "response"
        return json.loads(data)[key]



class Runnable:
    def __init__(s):
        s._title = None

    @property
    def title(s):
        """
        The title is used in :class:`Expyrimenter <expyrimenter.Executor>` log.
        """
        return s._title

    @title.setter
    def title(s, value):
        s._title = value

    # Exceptions will be displayed to the user
    def run(self):
        raise NotImplementedError('Runnable.run not implemented')



#!/usr/bin/env python3

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='expyrimenter',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # http://packaging.python.org/en/latest/tutorial.html#version
    version='0.2.0',

    description='Python tool to ease and parallelize distributed system'
                ' management tasks.',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/cemsbr/expyrimenter',

    # Author details
    author='Carlos Eduardo Moreira dos Santos',
    author_email='cems@cemshost.com.br',

    # Choose your license
    license='GPLv3',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Systems Administration',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    # What does your project relate to?
    keywords='shell ssh cluster cloud install setup parallel',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # List run-time dependencies here. These will be installed by pip when your
    # project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/technical.html#install-requires-vs-requirements-files
    install_requires=['requests'],

    # List additional groups of dependencies here (e.g. development
    # dependencies).
    # You can install these using the following syntax, for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
)



from os.path import join, expanduser, dirname
from configparser import ConfigParser
import logging


class Config:
    """
    ``Config.user_ini`` is the location of the user ini file and its default
    value is ``~/.expyrimenter/config.ini``.
    If you change the user config location,
    it will take effect in every new object.
    """
    user_ini = join(expanduser('~'), '.expyrimenter', 'config.ini')
    _default_ini = join(dirname(__file__), 'config.ini')
    _logger = logging.getLogger('config')

    def __init__(self, section):
        """
        :param str section: Ini file section name
        """
        parser = ConfigParser()
        parser.read([Config._default_ini, Config.user_ini])

        if section in parser.sections():
            self._section = parser[section]
        else:
            self._section = {}

    def get(self, key, default=None):
        """
        :param str key: The key whose value you are looking for
        :param any default: value to return if key is not found
        :return: *key* value
        :rtype: str
        """
        return self._section.get(key, default)
